# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.db_connection.oracle.connection import Oracle, OracleExtra
from onetl.connection.db_connection.oracle.dialect import OracleDialect
