# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from onetl.file.format.avro import Avro
from onetl.file.format.csv import CSV
from onetl.file.format.excel import Excel
from onetl.file.format.json import JSON
from onetl.file.format.jsonline import JSONLine
from onetl.file.format.orc import ORC
from onetl.file.format.parquet import Parquet
from onetl.file.format.xml import XML
