# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.file_df_connection.spark_s3.connection import SparkS3
from onetl.connection.file_df_connection.spark_s3.extra import SparkS3Extra
